/*
 * ---------------------------------------------------------------------------
 * This file defines constants, functions, ... used to control the lcd.
 * author : Mstl Georg
 * date   : 14.11.2006
 * ---------------------------------------------------------------------------
 */

 
#include "lcd_ctrl.h"
#include "gasmixer_defs.h"

#include <util/delay.h>	//_delay_ms(), _delay_us()
#include <string.h>     //strlen()


/* 
 * -------------------------------------------------------------------------
 * internal function declarations 
 * -------------------------------------------------------------------------
 */
 
void setPortDirOutLCD(void);
	
void setPortDirInLCD(void);
	
void setDataPortLowLCD(void);
	
void setCtrlPortLowLCD(void);
 
void pulseEnLCD(void);

uint8_t recByteLCD(uint8_t rs);
	
void sendByteLCD(uint8_t byte, uint8_t rs);
		
void waitBusyFlagLCD(void);	

	
/* 
 * -------------------------------------------------------------------------
 * internal function definitions
 * -------------------------------------------------------------------------
 */	
 
void setPortDirOutLCD(void) {

  //set direction of lcd ports to "out"								
	LCD_DDR |= 	( (1<<LCD_DATA_0) |
							  (1<<LCD_DATA_1) |
							  (1<<LCD_DATA_2) |
							  (1<<LCD_DATA_3) |
							  (1<<LCD_EN)     |
							  (1<<LCD_RS)     |
							  (1<<LCD_RW) );

}//setPortDirOutLCD


void setPortDirInLCD(void) {

  //set direction of lcd data ports to "in"				
	LCD_DDR &= ~ ( (1<<LCD_DATA_0) |
								 (1<<LCD_DATA_1) |
							   (1<<LCD_DATA_2) |
							   (1<<LCD_DATA_3) );
	
}//setPortDirInLCD


void setDataPortLowLCD(void) {

	//clear lcd ports -> set '0'
	LCD_PORT &= ~ ( (1<<LCD_DATA_0) |
							    (1<<LCD_DATA_1) |
							    (1<<LCD_DATA_2) |
							    (1<<LCD_DATA_3) );
	
}//setDataPortLowLCD


void setCtrlPortLowLCD(void) {

  //clear lcd ports -> set '0'
	LCD_PORT &= ~ ( (1<<LCD_EN)     |
									(1<<LCD_RS)     |
									(1<<LCD_RW) );

}//setCtrlPortLowLCD

 
void pulseEnLCD(void) {

	//set EN == 1 
	LCD_PORT |= (1<<LCD_EN);
	
	//wait min. pulse width of enable signal
	_delay_us(MIN_PULSE_WIDTH_EN_LCD);
	
	//set EN == 0 
	LCD_PORT &= ~ (1<<LCD_EN);
	
}//pulseEnLCD


uint8_t recByteLCD(uint8_t rs) {

  volatile uint8_t byte;
	uint8_t bits[8];
	uint8_t pos[] = { LCD_DATA_0, LCD_DATA_1, LCD_DATA_2, LCD_DATA_3 };
	uint8_t i;

  //set direction of data port to in and port values to 0 -> no external pull 
	//up resistor is used
	setDataPortLowLCD();
  setPortDirInLCD();
	
	//set "Register Select"
	//0 -> read cmd/adr; 1 -> read data
	if (rs) 
		LCD_PORT |= (1<<LCD_RS); 
  //else
		//LCD_PORT &= ~ (1<<LCD_RS);
	
	//set "Read/Write"
	//0 -> write; 1 -> read
	LCD_PORT |= (1<<LCD_RW);
	
	//read high nibble
	LCD_PORT |= (1<<LCD_EN);
	_delay_us(MIN_PULSE_WIDTH_EN_LCD);
	byte = LCD_PIN;
	LCD_PORT &= ~ (1<<LCD_EN);
		
	//determine 0 or 1 value for bit 0 - 3
	for(i=0; i<4; i++) 
		bits[i+4] = ( byte & (1<<pos[i]) ) >> pos[i];

	//read low nibble
	LCD_PORT |= (1<<LCD_EN);
	_delay_us(MIN_PULSE_WIDTH_EN_LCD);
	byte = LCD_PIN;
	LCD_PORT &= ~ (1<<LCD_EN);
	
	//determine 0 or 1 value for bit 0 - 3
	for(i=0; i<4; i++) 
		bits[i] = ( byte & (1<<pos[i]) ) >> pos[i];

  //determine received value -> transfer binary into decimal representation
  byte = 0;
	for(i=0; i<8; i++)
	  byte += bits[i]<<i;
		
  return byte;
  
}//recByteLCD


void sendByteLCD(uint8_t byte, uint8_t rs) {
	
	uint8_t bits[4] = {0};
	uint8_t i, j;

	waitBusyFlagLCD();
	
	//set port direction out and drive '0'
	setDataPortLowLCD();
	setCtrlPortLowLCD();
	setPortDirOutLCD();
	
	//set "Register Select"
	//0 -> write cmd/adr; 1 -> write data
	if(rs) 
		LCD_PORT |= (1<<LCD_RS);
  //else
		//LCD_PORT &= ~ (1<<LCD_RS);

	//set "Read/Write"
	//0 -> write; 1 -> read
	//LCD_PORT &= ~ (1<<LCD_RW);
		
	for(i=0; i<2; i++) {
			
		if(i==0) {
			
			//determine value 0 or 1 for bit 0 - 3 high nibble
			for(j=0; j<4; j++) 
				bits[j] = ( (byte>>4) & (1<<j) ) >> j;
	  
		} else {
		
			//determine low nibble
			for(j=0; j<4; j++) 
				bits[j] = ( (byte & 0x0F) & (1<<j) ) >> j;
		
		}//if
		
		//send nibble
		setDataPortLowLCD();
		LCD_PORT |= ( (bits[0]<<LCD_DATA_0) |
									(bits[1]<<LCD_DATA_1) |
									(bits[2]<<LCD_DATA_2) |
									(bits[3]<<LCD_DATA_3) );
		pulseEnLCD();
		
	}//for
		
}//sendByteLCD


void waitBusyFlagLCD() {

	//wait until busy flag is deasserted
	while(recByteLCD(0) & (1<<BUSY_FLAG_POS_LCD));
	
}//waitBusyFlag


/* 
 * -------------------------------------------------------------------------
 * general function definitions
 * -------------------------------------------------------------------------
 */	



void sendSignPosLCD(uint8_t sign, uint8_t pos) {

  //send address to lcd
	//"DD RAM Address Set" == b_1000_0000 + position
  sendByteLCD(DDR_ADR_SET + pos, 0);
	
	//send data to lcd
	sendByteLCD(sign, 1);

}//sendSignPosLCD


void sendTextPosLCD(char* text, uint8_t pos) {

  uint8_t i;
	uint8_t len = (uint8_t)strlen(text);
	
	for(i=0;i<len && i<MAX_LINE_LENGTH_LCD;i++){
		sendSignPosLCD(text[i], pos);
		pos++;
	}

}//sendTextPosLCD


void clearLCD(void) {

  //"Clear Display" == b_0000_0001
	sendByteLCD(CLEAR_DISPLAY, 0);
	
	//"Cursor At Home" == b_0000_0010
	sendByteLCD(CURSOR_AT_HOME, 0);

}//clearLCD


void initLCD(void) {

	//set all lcd ports to "out" and drive '0'
	//clear lcd ports -> set '0'
	setCtrlPortLowLCD();
	setDataPortLowLCD();
	setPortDirOutLCD();								
	
	
	//fulfill initialization routine as it is described in the HD44780 datasheet

  //wait max. time for "power supply rising" 
	_delay_ms(MIN_TIME_PWR_ON_LCD);
	
	//watch out that the busy flag can't be tested until after the lcd was 
	//initialized
	
	//initialize lcd first in 8 bit mode -> important so that we can change 
	//from every state in 4 bit mode
	//RS == RW == 0 
	
	//set "Function Set" cmd == b_0011_0000 (watch out that bit 3 - 0 are tied 
	//to ground; that's ok because they are not needed) -> has to be sent three 
	//times
	//DL == '1' == 8 bit interface
	LCD_PORT |= ( (1<<LCD_DATA_0) | 
								(1<<LCD_DATA_1) );										
	pulseEnLCD();
	_delay_ms(MIN_TIME_FIRST_BYTE_LCD);
	pulseEnLCD();
	_delay_ms(MIN_TIME_FIRST_BYTE_LCD);
	pulseEnLCD();
	
	
	//set 4 bit mode
	//"Function Set" == b_0010_0000
	setDataPortLowLCD();
	LCD_PORT |= (1<<LCD_DATA_1);
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);
	
	
	//set 4 bit, 2 lines, 5x7 font
	//"Function Set" == b_0010_1000
	//send high nibble
	setDataPortLowLCD();
	LCD_PORT |= (1<<LCD_DATA_1);
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);

	//send low nibble
	setDataPortLowLCD();
	LCD_PORT |= (1<<LCD_DATA_3);
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);
	
	
	//display on, cursor off, cursor blink off
	//"Display On/Off Control" == b_0000_1000
	//send high nibble
	setDataPortLowLCD();							
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);

	//send low nibble
	setDataPortLowLCD();
	LCD_PORT |= (1<<LCD_DATA_3);
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);
	
	
	//clear display
	//"Clear Display" == b_0000_0001
	//send high nibble
	setDataPortLowLCD();							
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);

	//send low nibble
	setDataPortLowLCD();
	LCD_PORT |= (1<<LCD_DATA_0);
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);
	
	
	//cursor auto increment, no display shift
	//"Entry Mode Set" == b_0000_0110
	//send high nibble
	setDataPortLowLCD();						
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);

	//send low nibble
	setDataPortLowLCD();
	LCD_PORT |= ( (1<<LCD_DATA_1) |
	              (1<<LCD_DATA_2) );
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);
	
	
	//display on, cursor off, cursor blink off
	//"Display On/Off Control" == b_0000_1100 
	//send high nibble
	setDataPortLowLCD();							
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);

	//send low nibble
	setDataPortLowLCD();
	LCD_PORT |= ( (1<<LCD_DATA_2) |
	              (1<<LCD_DATA_3) );
	pulseEnLCD();
	_delay_ms(MIN_TIME_BETW_BYTE_LCD);
		
	
	//the display is now initialized and the busy flag can be used for further 
	//transmissions
	
}//initLCD
